'''
 
 Filter to implement the creation on a NEW layer where each color area
 is contoured using BRL
 ....................................................................
 
 Release 1.0   initial
 Release 1.10  Tin Tran Changed Tile calculation from 64 to 128 and it seemed to work okay for GIMP 2.10.8
        
 
 Flow implemented to get the final result:      ...detail the major steps
 
 1. save context values
 2. create a new image, same characteristics of the Target Image
 3. create a BG layer as a copy of the Target Layer
 4. stretch values both on Target and on Source to avoid empty extremes at 0 and 255
    using python_fu_Stretch_RGB_colours
 4. calculate Mean Values for R, G, B both for the stretched Target and the stretched Source 
 5. duplicate twice the stretched Target Layer
    the 3 Layers will be suffixed R, A, L
 6. calculate the ratios between the Means R,G,B of the Source and the Means R,G,B of the Target
 7. call an embedded ancillary function to modify pixel by pixel on TargetLayerR using the ratios
 8. call an embedded ancillary function to modify pixel by pixel on TargetLayerA using the deltas
 9. convert the ratios to gamma values and produce TargetLayerL using the gammas
10. add a new layer TargetLayerI from the TargetLayerR with inverted values
11. set layer I to SOFTLIGHT_MODE and opacity to 22.2%; adjust R and A opacities to 33.3% 
12. create a single EFFECTS layer merging the four existing layers one at the time
13. create a Saturated Layer by dup the EFFECTS layer
14. calculate the saturation values of the Saturated layer and the Source layer
    then calculate the ratio and loop to apply the saturation modification
15. merge down again and re-apply python_fu ancillary filter Stretch_RGB_colours    
 
 ----------------------------------------------------------------
 COPYRIGHT NOTICE
 ----------------------------------------------------------------
 
 This program is free software you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation either version 3 of the License, or
 (at your option) any later version.  
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program if not, you can view the GNU General Public
 License version 3 at the web site http://www.gnu.org/licenses/gpl-3.0.html
 Alternatively you can write to the Free Software Foundation, Inc., 675 Mass
 Ave, Cambridge, MA 02139, USA.
 
 -------------------------------------------------------------------
 other info
 prerequisite: my ancillary filter in python "Stretch_RGB_colours"
               available under Gimp Scripts & Plug-ins
 -------------------------------------------------------------------
 
'''
#!/usr/bin/env python

from gimpfu import *
import time
import datetime

############################################################################
# include, then use: erro_msg or info_msg
#---------------------------------------------------------------------------
def erro_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(0) # box
    gimp.message ("ERROR: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
def info_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # error console
    gimp.message ("INFO: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
############################################################################

# ===================================================================
# SERVICE ROUTINE
# ------------------------------------------------------------------- 
# Find and return the average colour of a range of values in an image  
# ------------------------------------------------------------------- 

def modify_color (img, layer, method, REDval, GREENval, BLUEval) :
    ''' 
    
    Parameters:
                img : image The current image.
                layer : layer The layer of the image that is selected.
                REDval : ratio to modify RED values
                GREENval : ratio to modify GREEN values
                BLUEval : ratio to modify BLUE values
    '''
    # Indicates that the process has started.
    gimp.progress_init("Modifying the 3 colour components in " + layer.name + "...")

    # Set up an undo group, so the operation will be undone in one step.
    pdb.gimp_image_undo_group_start(img)
    
    # Get the layer position.
    pos = 0;
    for i in range(len(img.layers)):
        if(img.layers[i] == layer):
            pos = i
    
    # Create a new layer to save the results (otherwise is not possible to undo the operation).
    newLayer = gimp.Layer(img, layer.name + " temp", layer.width, layer.height, layer.type, layer.opacity, layer.mode)
    img.add_layer(newLayer, pos)
    layerName = layer.name
    
    # Clear the new layer.
    pdb.gimp_edit_clear(newLayer)
    newLayer.flush()
    
    # READ the pixels.
    try:
        # Calculate the number of tiles.
        tn = int(layer.width / 128)
        if(layer.width % 128 > 0):
            tn += 1
        tm = int(layer.height / 128)
        if(layer.height % 128 > 0):
            tm += 1
        
        # Iterate over the tiles.
        for i in range(tn):
            for j in range(tm):
                # Update the progress bar.
                gimp.progress_update(float(i*tm + j) / float(tn*tm))
        
                # Get the tiles.

                
                srcTile = layer.get_tile(False, j, i)
                dstTile = newLayer.get_tile(False, j, i)
                # Iterate over the pixels of each tile.
                tile_width = srcTile.ewidth
                tile_height = srcTile.eheight
                
                for x in range(tile_width):
                    for y in range(tile_height):
                         
                        # Get the pixel and totalize his values if inside the requested range.
                        pixel = srcTile[x,y]
                        red = ord(pixel[0])
                        green = ord(pixel[1])
                        blue = ord(pixel[2])
                        gray = (red+green+blue)/3
                        
                        if method == "R":
                            newred = min(255, int(round(red*REDval)))
                            newgreen = min(255, int(round(green*GREENval)))
                            newblue = min(255, int(round(blue*BLUEval)))
                            
                        if method == "A":
                            newred = max(0, min(255, int(round(red+(REDval*gray/127)))))
                            newgreen = max(0, min(255, int(round(green+(GREENval*gray/127)))))
                            newblue = max(0, min(255, int(round(blue+(BLUEval*gray/127)))))
                            
                        res = chr(newred) + chr(newgreen) + chr(newblue)
                        
                        # If the image has an alpha channel (or any other channel) copy his values.
                        if(len(pixel) > 3):
                            for k in range(len(pixel)-3):
                                res += pixel[k+3]
                                
                        # Save the value in the result layer.
                        dstTile[x,y] = res
   
        # Update the new layer.
        newLayer.flush()
        newLayer.merge_shadow(True)
        newLayer.update(0, 0, newLayer.width, newLayer.height)
        
        # Remove the old layer.
        img.remove_layer(layer)
        
        # Change the name of the new layer (two layers can not have the same name).
        newLayer.name = layerName
    
    except Exception as err:
        gimp.message("Unexpected error: " + str(err))
    
    # Close the undo group.
    pdb.gimp_image_undo_group_end(img)
    # End progress.
    pdb.gimp_progress_end()
    
    return(newLayer)

# ============================================================
# MAIN PROGRAM
# ------------------------------------------------------------

def TransferColours (targetImage, targetLayer, sourceLayer) :
     
    # Indicates that the process has started.
    gimp.progress_init("Transferring colours to " + targetLayer.name + "...")

    width = targetLayer.width
    height = targetLayer.height
    type = RGB
    Tname = targetLayer.name
    if '.' in Tname:
        split = Tname.split(".", 1)
        Tname = split[0]
    Sname = sourceLayer.name
    if '.' in Sname:
        split = Sname.split(".", 1)
        Sname = split[0]
        
    pdb.gimp_context_push
    currFG = pdb.gimp_context_get_foreground()
    currBG = pdb.gimp_context_get_background()
    
    newImage = pdb.gimp_image_new(width, height, type)
    
    newImage.disable_undo()
    
    newTargetLayer = pdb.gimp_layer_new_from_drawable(targetLayer, newImage)     # background layer
    newTargetLayer.mode = NORMAL_MODE
    newTargetLayer.name = "Background New Image Layer"
    newTargetLayer.opacity = 100.0 
    newImage.add_layer(newTargetLayer, 0)
    
    sourceImage = sourceLayer.image
    # Stretching values both on Target and on Source to avoid empty extremes at 0 and 255
    targetLayerStretched = pdb.python_fu_Stretch_RGB_colours (newImage, newTargetLayer, 1, 1)
    sourceLayercopy = sourceLayer.copy()
    sourceImage.add_layer (sourceLayercopy, 0)
    sourceLayerStretched = pdb.python_fu_Stretch_RGB_colours (sourceImage, sourceLayercopy, 1, 1)
    
    (TargetMeanRED, SD, ME, PX, CN, PC) = pdb.gimp_histogram (targetLayerStretched, HISTOGRAM_RED, 0, 255)  # bug fixed, Stretched was missing
    #info_msg ("Target mean RED = "+str(TargetMeanRED))
    (TargetMeanGREEN, SD, ME, PX, CN, PC) = pdb.gimp_histogram (targetLayerStretched, HISTOGRAM_GREEN, 0, 255)  # bug fixed, Stretched was missing  
    #info_msg ("Target mean GREEN "+str(TargetMeanGREEN))
    (TargetMeanBLUE, SD, ME, PX, CN, PC) = pdb.gimp_histogram (targetLayerStretched, HISTOGRAM_BLUE, 0, 255)  # bug fixed, Stretched was missing  
    #info_msg ("Target mean BLUE = "+str(TargetMeanBLUE))
    
    
    (SourceMeanRED, SD, ME, PX, CN, PC) = pdb.gimp_histogram (sourceLayerStretched, HISTOGRAM_RED, 0, 255)  
    #info_msg ("Source mean RED = "+str(SourceMeanRED))
    (SourceMeanGREEN, SD, ME, PX, CN, PC) = pdb.gimp_histogram (sourceLayerStretched, HISTOGRAM_GREEN, 0, 255)  
    #info_msg ("Source mean GREEN "+str(SourceMeanGREEN))
    (SourceMeanBLUE, SD, ME, PX, CN, PC) = pdb.gimp_histogram (sourceLayerStretched, HISTOGRAM_BLUE, 0, 255)  
    #info_msg ("Source mean BLUE = "+str(SourceMeanBLUE))
    
    RatioMRED = float (SourceMeanRED*1.00/TargetMeanRED*1.00)
    RatioMGREEN = float (SourceMeanGREEN*1.00/TargetMeanGREEN*1.00)
    RatioMBLUE = float (SourceMeanBLUE*1.00/TargetMeanBLUE*1.00)
    #info_msg ("transfer Means ratios are: RED "+str(RatioMRED)+", "+str(RatioMGREEN)+", "+str(RatioMBLUE))
    
    newTargetLayerL = targetLayerStretched.copy()     # copy for use with Levels  # bug fixed, Stretched was missing  
    newImage.add_layer (newTargetLayerL, 1)
    newTargetLayerL.mode = NORMAL_MODE
    newTargetLayerL.name = "New Image Layer for Levels"
    newTargetLayerL.opacity = 100.0 
    
    newTargetLayerA = targetLayerStretched.copy()     # copy for use with Colour Additions  # bug fixed, Stretched was missing  
    newImage.add_layer (newTargetLayerA, 0)
    newTargetLayerA.mode = NORMAL_MODE
    newTargetLayerA.name = "New Image Layer for Colour Additions"
    newTargetLayerA.opacity = 100.0 
    
    #pdb.gimp_display_new (newImage)
    #return()
    
    # NEW! EQUALIZE based on TOTALS, so that the ratios reflect only the ratio of the colours between them
    TOTSourceMean = SourceMeanRED+SourceMeanGREEN+SourceMeanBLUE
    SourceQuotaRED = SourceMeanRED/TOTSourceMean
    SourceQuotaGREEN = SourceMeanGREEN/TOTSourceMean
    SourceQuotaBLUE = SourceMeanBLUE/TOTSourceMean
    TOTTargetMean = TargetMeanRED+TargetMeanGREEN+TargetMeanBLUE
    TargetQuotaRED = TargetMeanRED/TOTTargetMean
    TargetQuotaGREEN = TargetMeanGREEN/TOTTargetMean
    TargetQuotaBLUE = TargetMeanBLUE/TOTTargetMean
    
    # =================================================================================== TRANSFER RATIOS OF COLOURS
    #info_msg ("Quotas Source = "+str(SourceQuotaRED)+","+str(SourceQuotaGREEN)+","+str(SourceQuotaBLUE))
    #info_msg ("Quotas Target = "+str(TargetQuotaRED)+","+str(TargetQuotaGREEN)+","+str(TargetQuotaBLUE))
    
    RatioQRED = float (SourceQuotaRED*1.00/TargetQuotaRED*1.00)
    RatioQGREEN = float (SourceQuotaGREEN*1.00/TargetQuotaGREEN*1.00)
    RatioQBLUE = float (SourceQuotaBLUE*1.00/TargetQuotaBLUE*1.00)
    #info_msg ("transfer ratios are: RED "+str(RatioQRED)+", "+str(RatioQGREEN)+", "+str(RatioQBLUE))
    newTargetLayerR = modify_color (newImage, targetLayerStretched, "R", RatioQRED, RatioQGREEN, RatioQBLUE) # bug fixed, Stretched was missing  
    
    #pdb.gimp_display_new (newImage)
    #return()
    
    # =================================================================================== TRANSFER DIFFERENCES OF COLOURS
    
    AddRED = float ((SourceQuotaRED-TargetQuotaRED) * 256.0)
    AddGREEN = float ((SourceQuotaGREEN-TargetQuotaGREEN) * 256.0) 
    AddBLUE = float ((SourceQuotaBLUE-TargetQuotaBLUE) * 256.0) 
    #info_msg ("transfer deltas are: RED "+str(AddRED)+", "+str(AddGREEN)+", "+str(AddBLUE)+", ")
    newTargetLayerA = modify_color (newImage, newTargetLayerA, "A", AddRED, AddGREEN, AddBLUE)
    
    #pdb.gimp_display_new (newImage)
    #return()
    
    # =================================================================================== TRANSFER USING LEVELS
    # gamma based on ratios
    powerHV=0.45
    powerLV=2.22
    if RatioMRED >= 1.0:
        powerR=powerHV
    else:
        powerR=powerLV
    gammaR = max (0.1, min((RatioMRED**powerR),10))  
    pdb.gimp_levels (newTargetLayerL, HISTOGRAM_RED, 0, 255, gammaR, 0, 255)
    if RatioMGREEN >= 1.0:
        powerG=powerHV
    else:
        powerG=powerLV
    gammaG = max (0.1, min((RatioMGREEN**powerG),10))  
    pdb.gimp_levels (newTargetLayerL, HISTOGRAM_GREEN, 0, 255, gammaG, 0, 255)
    if RatioMBLUE >= 1.0:
        powerB=powerHV
    else:
        powerB=powerLV
    gammaB = max (0.1, min((RatioMBLUE**powerB),10))  
    pdb.gimp_levels (newTargetLayerL, HISTOGRAM_BLUE, 0, 255, gammaB, 0, 255)
    #info_msg ("transfer gammas are: RED,GREEN,BLUE "+
    #         str(max (0.1, min((RatioMRED**powerR),10)))+","+
    #         str(max (0.1, min((RatioMGREEN**powerG),10)))+", "+
    #         str(max (0.1, min((RatioMBLUE**powerB),10)))
    #        )
    #pdb.gimp_display_new (newImage)
    #return()
    
    # =========================================================== COMBINE OUTCOMES "R" and "A" with BASE LEVELS
    newTargetLayerR.mode = NORMAL_MODE
    newTargetLayerR.name = "New Image Layer for Transferring Colours Ratios"
    newTargetLayerR.opacity = 33.3
    
    newTargetLayerA.mode = NORMAL_MODE
    newTargetLayerA.name = "New Image Layer for Adding Colours"
    newTargetLayerA.opacity = 33.3
    
    # add inverted values to smooth and add local contrast
    newTargetLayerI = newTargetLayerR.copy()     # copy for use with inverted value
    newImage.add_layer (newTargetLayerI, 0)
    newTargetLayerI.mode = SOFTLIGHT_MODE
    newTargetLayerI.name = "New Image Layer for Inverted values"
    newTargetLayerI.opacity = 22.2
    pdb.plug_in_vinvert (newImage, newTargetLayerI)
    
    #pdb.gimp_display_new (newImage)
    #return()
    
    # ============================================================= MERGE THEN TRANSFER THE SATURATION 
    Effects = pdb.gimp_image_merge_down (newImage, newTargetLayerR, CLIP_TO_BOTTOM_LAYER)
    Effects = pdb.gimp_image_merge_down (newImage, newTargetLayerA, CLIP_TO_BOTTOM_LAYER)
    Effects = pdb.gimp_image_merge_down (newImage, newTargetLayerI, CLIP_TO_BOTTOM_LAYER)
    
    #pdb.gimp_display_new (newImage)
    #return()
    
    # Get Saturation Mean from both layers, the source and the resulting outcome up to now
    
    (TSatImage, NA1, NA2, NA3) = pdb.plug_in_decompose (newImage, Effects, "Saturation", True)
    TSatLayer = TSatImage.active_layer
    (TSatMean, SD, ME, PX, CN, PC) = pdb.gimp_histogram (TSatLayer, HISTOGRAM_VALUE, 0, 255)  
    #info_msg ("TSat mean = "+str(TSatMean))
    
    (SSatImage, NA1, NA2, NA3) = pdb.plug_in_decompose (newImage, sourceLayerStretched, "Saturation", True)
    SSatLayer = SSatImage.active_layer
    (SSatMean, SD, ME, PX, CN, PC) = pdb.gimp_histogram (SSatLayer, HISTOGRAM_VALUE, 0, 255)  
    #info_msg ("SSat mean = "+str(SSatMean))
    
    TransferSAT = float (SSatMean*1.00/TSatMean*1.00)
    #info_msg ("transfer ratio is : "+str(TransferSAT))
    
    #pdb.gimp_display_new (newImage)
    #return()
    
    # using mod saturation to transfer from source to target
    
    Saturated = Effects.copy()
    newImage.add_layer (Saturated, 0)
    Saturated.name = "Layer with transferred Saturation"
    Saturated.opacity = 50.0
    Saturated.mode = COLOR_MODE
    if TransferSAT >= 1:
        increase = (TransferSAT-1) * 100
        while increase > 0:
            if increase >= 100:
                pdb.gimp_hue_saturation (Saturated, 0, 0, 0, 100)
                increase = increase-100
            else:
                pdb.gimp_hue_saturation (Saturated, 0, 0, 0, increase)
                increase = 0
    if TransferSAT < 1:
        decrease = (1-TransferSAT) * -100       # bug fixed there was no 1-
        pdb.gimp_hue_saturation (Saturated, 0, 0, 0, decrease)
        decrease = 0
    
    #pdb.gimp_display_new (newImage)
    #return()
    
    pdb.plug_in_gauss (newImage, Saturated, 5.0, 5.0, 1)      # not large blur: use RLE
    
    pdb.gimp_image_remove_layer (sourceImage, sourceLayerStretched)      # remove work layer
    pdb.gimp_image_delete (TSatImage)
    pdb.gimp_image_delete (SSatImage)
    
    FinalLayer = pdb.gimp_image_merge_down (newImage, Saturated, CLIP_TO_BOTTOM_LAYER)
    FinalLayer.name = Tname+", with colours modified using "+Sname
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H-%M-%S')
    FinalLayer.name = st+" "+FinalLayer.name
    
    # added after having seen outcomes w/o values at the extremes
    FinalLayerStretched = pdb.python_fu_Stretch_RGB_colours (newImage, FinalLayer, 1, 1)
    
    #pdb.gimp_display_new (newImage)
    #return()
    
    # End progress.
    pdb.gimp_progress_end()
    pdb.gimp_context_set_foreground(currFG)
    pdb.gimp_context_set_background(currBG)
    pdb.gimp_context_pop
    
    pdb.gimp_display_new(newImage)
    
    newImage.enable_undo()
    
    return(FinalLayerStretched)
    #return()

register(
    "TransferColours",
    "transfer colours from a source image to a target image",
    "returns a new image as a duplicate of the target image with colours modified according to the source image.",
    "Diego Nassetti",
    "Open source (BSD 3-clause license)",
    "January 2017",
    "TransferColours",
    "RGB, RGB*",
    [
      (PF_IMAGE, "image", "Target image", None),
      (PF_DRAWABLE, "target", "Target drawable", None),
      (PF_DRAWABLE, "source", "Source drawable", None),
    ],
    [
      (PF_DRAWABLE, "odrawable", "Output drawable", None),
    ],
    TransferColours,
    menu="<Image>/Diego/Colours"
    )
main()
